"""Python environment module for Django."""
import os
import json
import importlib
from copy import deepcopy
import django
import django.views.decorators.http
from loguru import logger
from escape_cli.executor import execute
from escape_cli.static import METHODS_PATH


def _patch_transactions():
    """Patch to hijack transactions in the middleware."""
    original_fn = deepcopy(importlib.import_module)

    def new_fn(*args, **kwargs):
        res = original_fn(*args, **kwargs)
        for arg in args:
            if 'settings' in arg:
                logger.success(f'Patching file: {arg}')
                res.MIDDLEWARE += (
                    'escape_cli.middlewares.django.GetRequestAndResponseInformation'
                    ,)
        return res
    importlib.import_module = new_fn


def _patch_child_decorators(patched_fn, hijacked_methods):
    """Patch child decorators in django.views.decorators.http, in particuliar @require_POST, @require_GET and @require_safe."""
    original_fn = deepcopy(patched_fn)

    def new_fn(*args, **kwargs):
        with open(METHODS_PATH) as f:
            methods_dicts = json.load(f)
        for arg in args:
            methods_dicts[f'{arg.__module__}.{arg.__name__}'
                ] = hijacked_methods
        with open(METHODS_PATH, 'w+') as f:
            json.dump(methods_dicts, f)
        res = original_fn(*args, **kwargs)
        return res
    return new_fn


def _patch_parent_decorator():
    """Patch django.views.decorators.http.require_http_methods, which is the parent decorator of all @require_X decorators."""
    original_fn = deepcopy(django.views.decorators.http.require_http_methods)

    def new_fn(*args, **kwargs):
        res = original_fn(*args, **kwargs)
        res = _patch_child_decorators(res, *args)
        return res
    django.views.decorators.http.require_http_methods = new_fn


def django_patch(entrypoint, as_module):
    """The `filename` is the file entrypoint executed in the controlled Python environment."""
    _patch_transactions()
    _patch_parent_decorator()
    django.views.decorators.http.require_POST = _patch_child_decorators(django
        .views.decorators.http.require_POST, ['POST'])
    django.views.decorators.http.require_GET = _patch_child_decorators(django
        .views.decorators.http.require_GET, ['GET'])
    django.views.decorators.http.require_safe = _patch_child_decorators(django
        .views.decorators.http.require_GET, ['GET', 'HEAD'])
    os.environ['ESCAPE_ENDPOINTS_MAPPED'] = 'False'
    execute(entrypoint, as_module)



from escape_cli.static.constants import DISCOVER_NAMESPACE
from .django import django_patch
from .flask import flask_patch
from loguru import logger


def patch_app(config, entrypoint, as_module):
    logger.info(f'> {entrypoint}')
    if config[DISCOVER_NAMESPACE]['httpLib'].lower() == 'django':
        django_patch(entrypoint, as_module)
    elif config[DISCOVER_NAMESPACE]['httpLib'].lower() == 'flask':
        flask_patch(entrypoint, as_module)



"""Small util file for retrieving vsc infos."""
import subprocess
from loguru import logger


def get_vcs_infos():
    """Execute git command in a subprocess for retreving commit infos.



    Return a dict with versioning infos for git projects, and an empty dict otherwise

    """
    try:
        infos = subprocess.check_output(['git', 'log', '-n', '1',
            '--pretty=tformat:%H,%s,%aI,%ae']).strip().decode('ascii').split(
            ',')
        branch_name = subprocess.check_output(['git', 'rev-parse',
            '--abbrev-ref', 'HEAD']).decode('ascii')
        return {'commit_hash': infos[0], 'commit_msg': infos[1],
            'commit_date': infos[2], 'commit_author': infos[3],
            'current_branch': branch_name}
    except:
        logger.warning(
            'No VCS informations found. Assuming this is not a git repository')
        return {}



"""Main."""
import sys
import click
from loguru import logger
from . import __version__, commands
from .utils import title, format_logs, parse_command
from .static import OPTION_DICTIONNARY
logger.remove()
logger.add(sys.stderr, format=format_logs)
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.option('-V', '--version', 'version', default=False, is_flag=True)
def main(version):
    """Starting point of the Python CLI."""
    if version:
        print(__version__)
        sys.exit(0)
    title()


main.add_command(commands.init)
main.add_command(commands.discover)
main.add_command(commands.cover)
main.add_command(commands.scan)
if len(sys.argv) > 1 and sys.argv[1] in OPTION_DICTIONNARY.keys():
    main(parse_command(sys.argv))



"""Handle .escaperc config file."""
import os
import json
from loguru import logger
from escape_cli.static.constants import ESCAPE_DIR, TRANSACTIONS_PATH, ENDPOINTS_PATH, METHODS_PATH
from escape_cli.patchs import patch_app


def get_config(config_filename, discover_namespace):
    """Get configuration to correctly patch the application."""
    if not os.path.isfile(config_filename):
        logger.error(
            f'Config file at path {config_filename} not found. Did you run `escape-py init`?'
            )
        return None
    with open(config_filename, 'r') as f:
        config = json.load(f)
        f.close()
        if not isinstance(config, dict
            ) or discover_namespace not in config.keys():
            logger.error(
                f'Config file at path {config_filename} has a wrong format. Did you run `escape-py init`?'
                )
            return None
        return config


def init_logs(escape_dir, transactions_path, endpoints_path, methods_path):
    """Reset the escape config directory where logs will be saved."""
    if not os.path.isdir(escape_dir):
        os.mkdir(escape_dir)
    with open(transactions_path, 'w+') as f:
        f.write('[')
        logger.success(f'Transactions have been reset: {transactions_path}')
    with open(endpoints_path, 'w+') as f:
        logger.success(f'Endpoints have been reset: {endpoints_path}')
    with open(methods_path, 'w+') as f:
        json.dump({}, f)
        logger.success(f'Methods have been reset: {methods_path}')


def close_json(char, file):
    """Properly end the json file."""

    def move_backward():
        """Move one char backward in the file."""
        fb.seek(-1, os.SEEK_END)
    with open(file, 'rb+') as fb:
        move_backward()
        if fb.read(1) == b',':
            move_backward()
            fb.truncate()
        fb.write(bytes(char, 'utf-8'))


def close_logs(transactions_path, endpoints_path):
    """Once completed format logs to make them readable."""
    close_json(']', transactions_path)
    with open(transactions_path, 'r') as f:
        transactions = json.load(f)
    try:
        with open(endpoints_path, 'r') as f:
            patterns = json.load(f)
    except json.JSONDecodeError as error:
        raise ValueError(
            'Unable to parse the endpoints. You may need to make at least one request to the API.'
            ) from error
    return transactions, patterns


def patch_and_run(entrypoint, config, as_module, manual=False):
    """Patch the app and run the entrypoint."""
    result = {}
    init_logs(ESCAPE_DIR, TRANSACTIONS_PATH, ENDPOINTS_PATH, METHODS_PATH)
    try:
        patch_app(config, entrypoint, as_module)
    except KeyboardInterrupt as error:
        if manual:
            logger.warning(
                "KeyboardInterrupt exception detected, but the flag `--manual` is enabled,  so we'll send results anyway."
                )
        else:
            raise error
    result['transactions'], result['endpoints'] = close_logs(TRANSACTIONS_PATH,
        ENDPOINTS_PATH)
    return result



"""Utils to save patching results."""
import json
from loguru import logger
from escape_cli.static.constants import TRANSACTIONS_PATH


def save_transaction(result):
    """Save request and response informations in the transactions file."""
    for key in ['req', 'res', '_identifier']:
        if key not in result.keys():
            logger.warning(f'\n{key} is not in the result')
            return
    with open(TRANSACTIONS_PATH, 'a') as f:
        f.write(f'{json.dumps(dict(result))},')



"""Options shared between the commands."""
from os.path import isfile
from functools import wraps
import click
from escape_cli.static.constants import DISCOVER_NAMESPACE
from escape_cli.utils.config import get_config
from escape_cli.static.command_options import OPTION_DICTIONNARY


def extract_config(func):
    """Extract the config given the config path."""

    @wraps(func)
    def wrapped(*args, **kwargs):
        config_path = kwargs['config_path']
        config = get_config(config_path, DISCOVER_NAMESPACE)
        if not config:
            raise ValueError(
                f'Cannot read the config in the file {config_path}')
        del kwargs['config_path']
        kwargs['config'] = config
        return func(*args, **kwargs)
    return wrapped


def extract_as_module(func):
    """Check if the entrypoint correspond to a file in the cwd."""

    @wraps(func)
    def wrapped(*args, **kwargs):
        kwargs['as_module'] = not isfile(kwargs['entrypoint'][0].split(' ')[0])
        return func(*args, **kwargs)
    return wrapped


def add_options(func):
    """Add click options."""

    def add_option(func, command_option):
        """Add click option."""
        return click.option(*command_option['options'], command_option[
            'name'], default=command_option.get('default'), help=
            command_option.get('default', ''), is_flag=command_option.get(
            'is_flag'))(func)
    command_options = OPTION_DICTIONNARY[func.__name__]
    func_with_options = func
    for option in command_options:
        func_with_options = add_option(func_with_options, option)
    return func_with_options


def parse_command(argv):
    """Separate the options from the escape command and the app entrypoint."""
    command = argv[1]
    index = 2
    for option in OPTION_DICTIONNARY[command]:
        index += 1 if option.get('is_flag') else 2
    return argv[1:index] + [' '.join(argv[index:])]



"""Exec Python file in controlled environment.



Credits: Codecov

https://github.com/nedbat/coveragepy

"""
import os
import sys
import inspect
import platform
from types import ModuleType
from importlib import machinery
from importlib.util import find_spec
from multiprocessing import Process, set_start_method
from loguru import logger
from coverage.execfile import make_code_from_py, make_code_from_pyc


class DummyLoader:
    """A shim for the pep302 __loader__, emulating pkgutil.ImpLoader.



    Currently only implements the .fullname attribute

    """

    def __init__(self, fullname, *_args):
        self.fullname = fullname


class NoSource(Exception):
    """We couldn't find the source for a module."""


def find_module(modulename):
    """Find the module named `modulename`.



    Returns the file path of the module, the name of the enclosing package, and the spec.

    """
    try:
        spec = find_spec(modulename)
    except ImportError as err:
        raise NoSource(str(err)) from err
    if not spec:
        raise NoSource(f'No module named {modulename}')
    origin = spec.origin
    packagename = spec.name
    if spec.submodule_search_locations:
        mod_main = modulename + '.__main__'
        spec = find_spec(mod_main)
        if not spec:
            raise NoSource(
                f'No module named {mod_main}; {modulename} is a package and cannot be directly executed'
                )
        origin = spec.origin
        packagename = spec.name
    packagename = packagename.rpartition('.')[0]
    pathname = str(origin)
    return pathname, packagename, spec


def set_sys_path(args, as_module):
    """Set sys.path properly.



    This needs to happen before any importing, and without importing anything.

    """
    cpython = platform.python_implementation() == 'CPython'
    pyversion = sys.version_info + (int(platform.python_version()[-1] == '+'),)
    actual_syspath0_dash_m = cpython and pyversion >= (3, 7, 0, 'beta', 3)
    if as_module:
        if actual_syspath0_dash_m:
            path0 = os.getcwd()
        else:
            path0 = ''
    elif os.path.isdir(args[0]):
        path0 = args[0]
    else:
        path0 = os.path.abspath(os.path.dirname(args[0]))
    if os.path.isdir(sys.path[0]):
        top_file = inspect.stack()[-1][0].f_code.co_filename
        sys_path_0_abs = os.path.abspath(sys.path[0])
        top_file_dir_abs = os.path.abspath(os.path.dirname(top_file))
        if sys_path_0_abs != top_file_dir_abs:
            path0 = ''
    elif sys.path[1] == os.path.dirname(sys.path[0]):
        del sys.path[1]
    if path0:
        sys.path[0] = os.path.abspath(path0)


def prepare(args, as_module):
    """Do more preparation to run Python code.



    Includes finding the module to run and adjusting sys.argv[0]. This method is allowed to import code.

    """
    package = ''
    spec = None
    if as_module:
        modulename = args[0]
        pathname, package, spec = find_module(modulename)
        if spec:
            modulename = spec.name
        loader = DummyLoader(modulename)
        args[0] = os.path.abspath(pathname)
    elif os.path.isdir(args[0]):
        for ext in ['.py', '.pyc', '.pyo']:
            try_filename = os.path.join(args[0], '__main__' + ext)
            if os.path.exists(try_filename):
                args[0] = try_filename
                break
        else:
            raise NoSource(f'Can\'t find "__main__" module in "{args[0]}"')
        try_filename = os.path.abspath(args[0])
        spec = machinery.ModuleSpec('__main__', None, origin=try_filename)
        spec.has_location = True
        loader = DummyLoader('__main__')
    else:
        loader = DummyLoader('__main__')
    args[0] = os.path.abspath(args[0])
    return args, package, spec, loader


def execute(entrypoint, as_module):
    """The `entrypoint` is the entrypoint to be executed in python (with parameters)"""
    args = entrypoint.split()
    set_sys_path(args, as_module)
    args, package, spec, loader = prepare(args, as_module)
    env_mod = ModuleType('__main__')
    env_mod.__loader__ = loader
    env_mod.__file__ = args[0]
    from_pyc = args[0].endswith(('.pyc', '.pyo'))
    if from_pyc:
        env_mod.__file__ = env_mod.__file__[:-1]
    if package:
        env_mod.__package__ = package
    if spec:
        env_mod.__spec__ = spec
    if from_pyc:
        code = make_code_from_pyc(args[0])
    else:
        code = make_code_from_py(args[0])
    os.getcwd()
    sys.argv = args
    sys.modules['__main__'] = env_mod
    set_start_method('fork')
    execution = Process(target=exec, args=(code, env_mod.__dict__))
    execution.start()
    execution.join()
    if execution.exitcode:
        raise Exception(
            f'Unit test failed with exit code : {execution.exitcode}')
    logger.success('Unit tests succeeded')



"""Options shared between the commands."""
from escape_cli.static.constants import CONFIG_FILENAME
config_option = {'name': 'config_path', 'options': ['-c', '--config'],
    'default': CONFIG_FILENAME, 'help': 'Path of the escape config file.'}
manual_option = {'name': 'manual', 'options': ['--manual'], 'is_flag': True}
OPTION_DICTIONNARY = {'init': [config_option], 'cover': [config_option],
    'discover': [config_option, manual_option], 'scan': [config_option]}



"""Init."""
import json
import os.path
from pathlib import Path
from importlib import resources
import click
import inquirer
from loguru import logger
from escape_cli import static
from escape_cli.utils.commands import add_options


@click.command()
@add_options
@logger.catch
def init(config_path):
    """Init Escape Python Instrumenter by creating escaperc file."""
    with resources.open_text(static, 'supported-libs.json') as f:
        supported_libs = json.load(f)
    project_questions = [inquirer.Text(name='key', message=
        'Please enter the Escape project key')]
    project_answers = inquirer.prompt(project_questions,
        raise_keyboard_interrupt=True)
    answers = {'project': project_answers}
    discover_questions = [inquirer.List(name='httpLib', choices=
        supported_libs['httpLib'].keys(), message=
        'What HTTP library do you use for this project?')]
    discover_answers = inquirer.prompt(discover_questions,
        raise_keyboard_interrupt=True)
    answers.update({'discover': discover_answers})
    logged_action = 'written'
    if os.path.isfile(config_path):
        with open(config_path) as f:
            existing_config = json.load(f)
            for key, val in existing_config.items():
                if key not in answers:
                    answers[key] = val
            logged_action = 'updated'
    with open(config_path, 'w') as f:
        json.dump(answers, f, indent=2)
        logger.success(
            f'Config file {logged_action}: {Path(config_path).resolve()}')



"""Escape Intelligence SDK used to communicate with Escape Backend using project tokens. This SDK have limited capabilities:



- Edit a project

- Create, update and add observed transactions to runs

- Create alerts

"""
import os
import json
import socket
import requests
from loguru import logger
from escape_cli.static.constants import INFOS_PATH
from escape_cli.utils.http import prepare_request_data, parse_dates, serialize_dates
from escape_cli.utils.vcs import get_vcs_infos
from escape_cli import __version__ as cli_version
HEADERS = {'content-type': 'application/json'}


def raise_for_status(response):
    """Raise error when HTTP code is not 200 OK."""
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        logger.error(f"{e}. {response.json().get('data', '')}")


@logger.catch
class EscapeIntelligenceSDK:
    """Escape API SDK dedicated to use in intelligence contexts.



    Authenticates using project Key.

    """

    def __init__(self, uri=''):
        """Creates a new IntelligenceSDK using a project URI

        ex: uri='https://app.escape.tech/2865e0e37546e069bfa65b06d74b6e8c2b0650d5'

        TODO: Use urllib.parse.urlparse here

        """
        protocol, address = uri.split('://')
        hostname, project_key = address.split('/')
        self.base_url = f'{protocol}://{hostname}'
        self.project_key = project_key

    def update_project(self, project_data, options=None):
        """Update the related project using the `project_data` given as argument."""
        options = options or {}
        response = requests.put(
            f'{self.base_url}/intelligence/{self.project_key}/project',
            headers=HEADERS, data=prepare_request_data(project_data), **options
            )
        raise_for_status(response)
        return response.json(object_hook=parse_dates)

    def create_run(self, mode, options=None):
        """Create a run with `run_data` that will be automatically linked to current project."""
        options = options or {}
        run_data = {'hostname': socket.gethostname(), 'status':
            'running:discovery', 'mode': mode}
        run_data.update(get_vcs_infos())
        run_data.update({'cli_version': cli_version})
        response = requests.post(
            f'{self.base_url}/intelligence/{self.project_key}/runs',
            headers=HEADERS, data=prepare_request_data(run_data), **options)
        raise_for_status(response)
        run = response.json(object_hook=parse_dates)
        directory = os.path.dirname(INFOS_PATH)
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(INFOS_PATH, 'w') as f:
            json.dump(run, f, default=serialize_dates)
            logger.success(f'Run infos saved: {INFOS_PATH}')
        return response.json(object_hook=parse_dates)

    def get_run(self, uuid, options=None):
        """Get a run using it's `uuid`."""
        options = options or {}
        response = requests.get(
            f'{self.base_url}/intelligence/{self.project_key}/runs/{uuid}',
            headers=HEADERS, **options)
        raise_for_status(response)
        return response.json(object_hook=parse_dates)

    def add_transactions_to_run(self, uuid, transactions, options=None):
        """Add a `transaction` to a run using it's `uuid`."""
        options = options or {}
        response = requests.post(
            f'{self.base_url}/intelligence/{self.project_key}/runs/{uuid}/transactions'
            , headers=HEADERS, data=prepare_request_data({'transactions':
            transactions}), **options)
        raise_for_status(response)
        return response.json(object_hook=parse_dates)

    def add_endpoints_to_run(self, uuid, endpoints, options=None):
        """Add `enpoints` to a run using it's `uuid`."""
        options = options or {}
        response = requests.post(
            f'{self.base_url}/intelligence/{self.project_key}/runs/{uuid}/endpoints'
            , headers=HEADERS, data=prepare_request_data({'endpoints':
            endpoints}), **options)
        raise_for_status(response)
        return response.json(object_hook=parse_dates)

    def add_alerts_to_run(self, uuid, alert_data, options=None):
        """Add an `alert` to a run using it's `uuid`."""
        options = options or {}
        response = requests.post(
            f'{self.base_url}/intelligence/{self.project_key}/runs/{uuid}/alerts'
            , headers=HEADERS, data=prepare_request_data(alert_data), **options
            )
        raise_for_status(response)
        return response.json(object_hook=parse_dates)

    def generate_run_openapi_spec(self, uuid, options=None):
        """Generate and return the OpenAPI specification for the run identified by its `uuid` using the transactions it contains."""
        options = options or {}
        response = requests.patch(
            f'{self.base_url}/intelligence/{self.project_key}/runs/{uuid}/generate-openapi'
            , headers=HEADERS, **options)
        raise_for_status(response)
        return response.json(object_hook=parse_dates)

    def update_run(self, uuid, run_data, options=None):
        """Update the related project using the `project_data` given as argument."""
        options = options or {}
        response = requests.put(
            f'{self.base_url}/intelligence/{self.project_key}/runs/{uuid}',
            headers=HEADERS, data=prepare_request_data(run_data), **options)
        raise_for_status(response)
        return response.json(object_hook=parse_dates)



"""Discovery."""
import sys
import traceback
import click
from loguru import logger
from termcolor import colored
import escape_cli.utils.coverage as coverage
from escape_cli.utils.sdk import EscapeIntelligenceSDK
from escape_cli.static.constants import COVERAGE_NAMESPACE, DISCOVER_NAMESPACE, PROJECT_NAMESPACE
from escape_cli.utils.config import patch_and_run
from escape_cli.utils.commands import add_options, extract_as_module, extract_config


@click.command()
@click.argument('entrypoint', nargs=-1, required=True, type=str)
@add_options
@extract_config
@extract_as_module
@logger.catch
def discover(entrypoint, config, manual, as_module):
    """ Discover routes on your application while running your tests and assess coverage. This will send results to the Escape web application.



    

    Example: escape-py discover pytest

    Example: escape-py discover --config .escaperc.staging.json pytest

    """
    project_uri = config[PROJECT_NAMESPACE]['key']
    coverage_config = config[DISCOVER_NAMESPACE].get(COVERAGE_NAMESPACE, {})
    if manual:
        logger.warning('')
        logger.warning(colored('MANUAL MODE DETECTED', 'yellow', attrs=[
            'bold']))
        logger.warning(colored('Discovery results will be stored ',
            'yellow') + colored('regardless of the exit code', 'yellow',
            attrs=['bold']) + colored(
            ', including at keyboard interruption', 'yellow'))
        logger.warning(colored('To prevent this, ', 'yellow') + colored(
            'hit CTRL+C twice for exiting', 'yellow', attrs=['bold']) +
            colored(' or ', 'yellow') + colored(
            'run without the manual mode', 'yellow', attrs=['bold']))
        logger.warning('')
    logger.info('Creating run')
    client = EscapeIntelligenceSDK(project_uri)
    run = client.create_run('manual' if manual else 'automatic')
    logger.info(f"Run created with uuid {run['uuid']}")
    try:
        result = patch_and_run(' '.join(entrypoint), config, as_module, manual)
        if not result:
            return
        coverage_data = coverage.run_coverage(result['transactions'],
            result['endpoints'], coverage_config)
        filtered_coverage, coverage_stats, enriched_endpoints = coverage_data
        logger.info('Saving HTTP messages into the DB')
        client.add_transactions_to_run(run['uuid'], transactions=result[
            'transactions'])
        logger.info('Generating OpenAPI specification')
        client.generate_run_openapi_spec(run['uuid'])
        logger.info('Saving the endpoints in the DB')
        client.add_endpoints_to_run(run['uuid'], endpoints=enriched_endpoints)
        logger.info('Saving run metadata')
        client.update_run(run['uuid'], {'status': 'success:discovery', **
            coverage_stats})
        coverage.display_coverage_reports(filtered_coverage, coverage_stats,
            coverage_config)
        logger.success('Succesfully discovered routes on your application')
        logger.info(
            f"To proceed with scan, run `escape-py scan start --run-id {run['uuid']}` in your project's folder"
            )
    except Exception as err:
        traceback.print_exc()
        logger.error(err)
        client.update_run(run['uuid'], {'status': 'failed:discovery'})
        sys.exit(1)



"""Main."""
import json
from flask import request, Response, Flask
from loguru import logger
from escape_cli.utils.result import save_transaction
from escape_cli.utils.parse import find_params, parse_parameters, format_lib_to_openapi_path, get_identifier
from escape_cli.static.constants import ENDPOINTS_PATH


def decode_res_body(response):
    """Decode the body response."""
    res_body = {}
    if response.is_json:
        res_body = cast(dict, response.get_json())
    else:
        try:
            res_body = cast(str, response.data.decode('utf-8'))
        except UnicodeDecodeError:
            logger.warning('Enable to decode the body. It is set to empty.')
    return res_body


def flask_middleware(app):
    """Add a middleware called after each request."""

    @app.before_first_request
    def detect_routes():
        """Detect registered routes."""
        endpoints = []
        existing_routes = []
        for rule in app.url_map.iter_rules():
            lib_path = str(rule)
            handler = rule.endpoint
            parameters = find_params(lib_path)
            openapi_path = format_lib_to_openapi_path(lib_path)
            if openapi_path not in existing_routes:
                existing_routes.append(openapi_path)
                for method in (rule.methods - {'HEAD', 'OPTIONS'}):
                    endpoints.append({'openApiPath': openapi_path, 'method':
                        method, 'parameters': parameters, 'handler':
                        handler, '_identifier': get_identifier(method,
                        openapi_path)})
        with open(ENDPOINTS_PATH, 'w') as f:
            json.dump(endpoints, f)

    @app.after_request
    def detect_transaction(response):
        """Detect transactions."""
        save_req_and_res_information(response)
        return response


def save_req_and_res_information(response):
    """Fetch usefull information from requests and responses Save it in the transactions file."""
    response.direct_passthrough = False
    req_parameters = parse_parameters(request.view_args
        ) if request.view_args else {}
    openapi_path = format_lib_to_openapi_path(str(request.url_rule))
    result = {'req': {'openApiPath': openapi_path, 'protocol': request.
        scheme, 'host': request.host, 'route': request.path, 'method':
        request.method, 'parameters': req_parameters, 'query': dict(request
        .args), 'originalUrl': request.path, 'headers': dict(request.
        headers), 'cookies': dict(request.cookies), 'httpVersion': '',
        'body': dict(request.form)}, 'res': {'statusCode': response.
        status_code, 'messageCode': response.status.split(' ')[1],
        'headers': dict(response.headers), 'body': decode_res_body(response
        )}, '_identifier': get_identifier(request.method, openapi_path)}
    save_transaction(result)



"""Utils to manipulate HTTP Requests."""
import datetime
from loguru import logger
from simplejson import dumps


def parse_dates(dct, date_format='%Y-%m-%dT%H:%M:%S.%fZ'):
    """Parses string dates of a `dict` to a Python datetime `date_format`."""
    for key, value in dct.items():
        if isinstance(value, str) and value.endswith('Z'):
            try:
                dct[key] = datetime.datetime.strptime(value, date_format)
            except ValueError as e:
                logger.warning(f'Coul not parse date "{value}": {e}')
    return dct


def serialize_dates(obj):
    """Serializes dates in `obj` to backend-readable date strings."""
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    raise TypeError('Type %s is not serializable' % type(obj))


def prepare_request_data(data):
    """Clean `data` dictionnary and format it for to send with Python `requests` module.



    Python dicts might contains some non-json-serializable data such as NaNs which we must remove before sending to backend.

    """
    if isinstance(data, dict):
        data = {k: data[k] for k in data if not isinstance(k, (int, float))}
        data = dumps(data, default=serialize_dates, ignore_nan=True)
    return data.replace('\\u0000', '')



"""Python environment module for Flask."""
import flask
from escape_cli.executor import execute
from escape_cli.middlewares import flask_middleware


def flask_patch(entrypoint, as_module):
    """The `filename` is the file entrypoint executed in the controlled Python environment."""


    class NewClass(flask.Flask):
        """Patched Flask class."""

        def __init__(self, *args, **kwargs):
            """Overriden constructor."""
            super().__init__(*args, **kwargs)
            flask_middleware(self)
    flask.Flask = NewClass
    execute(entrypoint, as_module)



"""Init."""
import click
from loguru import logger
import escape_scanner_wrapper


@click.command()
@click.argument('options', nargs=-1, required=False, type=str)
@logger.catch
def scan(options):
    """Start the scanner and pass the options."""
    escape_scanner_wrapper.execute(*options)



"""Cover."""
import sys
import click
from loguru import logger
from escape_cli.static.constants import COVERAGE_NAMESPACE, DISCOVER_NAMESPACE
import escape_cli.utils.coverage as coverage
from escape_cli.utils.config import patch_and_run
from escape_cli.utils.commands import add_options, extract_as_module, extract_config


@click.command()
@click.argument('entrypoint', nargs=-1, required=True, type=str)
@add_options
@extract_as_module
@extract_config
@logger.catch
def cover(entrypoint, config, as_module):
    """ Discover routes on your application while running your tests and assess routes coverage. This will NOT send results to the Escape web application.



    

    Example: escape-py cover pytest

    Example: escape-py cover --config .escaperc.staging.json pytest

    """
    coverage_config = config[DISCOVER_NAMESPACE].get(COVERAGE_NAMESPACE, {})
    result = patch_and_run(' '.join(entrypoint), config, as_module)
    if not result:
        return
    coverage_data = coverage.run_coverage(result['transactions'], result[
        'endpoints'], coverage_config)
    filtered_coverage, coverage_stats, _ = coverage_data
    try:
        coverage.display_coverage_reports(filtered_coverage, coverage_stats,
            coverage_config)
    except ValueError as err:
        logger.error(err)
        sys.exit(1)



"""Constants."""
import os
CONFIG_FILENAME = '.escaperc.json'
PROJECT_NAMESPACE = 'project'
DISCOVER_NAMESPACE = 'discover'
COVERAGE_NAMESPACE = 'coverage'
ESCAPE_DIR = os.path.join(os.path.expanduser('~'), '.config', 'escape')
TRANSACTIONS_PATH = os.path.join(ESCAPE_DIR, 'transactions.json')
INFOS_PATH = os.path.join(ESCAPE_DIR, 'infos.json')
ENDPOINTS_PATH = os.path.join(ESCAPE_DIR, 'endpoints.json')
METHODS_PATH = os.path.join(ESCAPE_DIR, 'methods.json')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['escape_cli',
 'escape_cli.commands',
 'escape_cli.executor',
 'escape_cli.helpers',
 'escape_cli.middlewares',
 'escape_cli.patchs',
 'escape_cli.static',
 'escape_cli.utils']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2,<4.0',
 'Flask>=1.1.2,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'coverage>=5.5,<6.0',
 'inquirer>=2.7.0,<3.0.0',
 'loguru>=0.5.3,<0.6.0',
 'requests>=2.25.1,<3.0.0',
 'simplejson>=3.17.2,<4.0.0',
 'termcolor>=1.1.0,<2.0.0']

extras_require = \
{':sys_platform == "darwin"': ['escape-scanner-darwin-x64>=1.0.0-beta.21,<2.0.0'],
 ':sys_platform == "linux"': ['escape-scanner-linux-x64>=1.0.0-beta.21,<2.0.0']}

entry_points = \
{'console_scripts': ['escape-py = escape_cli:main']}

setup_kwargs = {
    'name': 'escape-py',
    'version': '1.0.0b21',
    'description': "Escape's Python CLI",
    'long_description': "# Python CLI\n\nEscape's Command Line Interface for Python\n\n## Install for development\n\n### Install Poetry globally\n\nPoetry is a `pip` replacement which is very close to `npm` in Node.js.\n\n`python3.9 -m pip install poetry`\n\n### Set your Poetry default venv path\n\nMore info: https://python-poetry.org/docs/configuration/#cache-dir\n\nYou can either use a global Virtual Env for all your Python projects such as:\n`poetry config virtualenvs.path ~/.venv`\n\nOr create a Virtual Env specific to your project:\n`poetry config virtualenvs.in-project true`\n\n### Main Functions\n\nMore info: https://python-poetry.org/docs/cli/\n\nBasic functions:\n\n- `poetry shell`: Source the Virtual Env of the repo\n- `poetry install`: Install all dependencies\n- `poetry install --no-dev`: Install all dependencies expect dev dependencies\n- `poetry update`: Update dependencies\n- `poetry add my-depency`: Add dependency\n\n## Install Escape-CLI in editable mode in another project\n\nTo install the CLI in the application Virtual Env on which you want to use the CLI in editable mode (meaning you will not need to reinstall it a each edit), simply run from the root of this directory:\n\n`./scripts/install-in-venv.sh ../../path/to/my/.venv/bin/activate`\n\nNote the path can be relative or absolute.\n",
    'author': 'Escape Technologies SAS',
    'author_email': 'ping@escape.tech',
    'maintainer': 'Antoine Carossio',
    'maintainer_email': 'antoine.carossio@me.com',
    'url': 'https://escape.tech/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Parsing modules."""
import re
import json
from loguru import logger


def parse_parameters(raw_params):
    """Parse potential uuid values in parameters."""
    parameters = {}
    for key in raw_params:
        value = raw_params[key]
        if not isinstance(value, str):
            try:
                parameters[key] = str(value)
                continue
            except UnicodeDecodeError:
                logger.error(
                    f'Param in url cannot be converted into string: {key}: {value}'
                    )
        parameters[key] = value
    return parameters


def parse_body(raw_body, content_type):
    """Parse body into json when content_type is application/json."""
    if not content_type:
        return {}
    decoded_body = raw_body.decode('utf-8') if isinstance(raw_body, bytes
        ) else raw_body
    if 'application/json' in content_type:
        try:
            parsed_body = json.loads(decoded_body)
            return parsed_body
        except json.decoder.JSONDecodeError:
            return 'Invalid JSON: ' + decoded_body
    else:
        return decoded_body


def interpolate_openapi_path(raw_route, params):
    """Interpolate the raw_route to Open API path using the parameters in params."""
    openapi_path = raw_route
    for key, value in params.items():
        openapi_path = openapi_path.replace(value, f'{{{key}}}')
    openapi_path = f"/{openapi_path.strip('$^/')}"
    return openapi_path


def format_lib_to_openapi_path(raw_path):
    """Format the path parameters according to the Open API Spec."""
    param_patterns = ['\\^?\\(\\?P\\<[^>]+\\>[^)]*\\)', '\\<[^>]+\\>']
    openapi_path = raw_path
    for regex in param_patterns:
        for raw_param in re.findall(regex, openapi_path):
            param = re.findall('\\<(?:.+:)?([^>]+)\\>', raw_param)[0]
            openapi_path = openapi_path.replace(raw_param, '{' + param + '}')
    openapi_path = f"/{openapi_path.strip('$^/')}"
    return openapi_path


def find_params(path, regex=''):
    """Find name, type and pattern of params in path."""
    parameters = {}
    for raw_param in re.findall('\\<(.*?)\\>', path):
        splitted_params = raw_param.split(':')
        name = splitted_params[-1]
        new_param = {'name': name}
        if len(splitted_params) == 2:
            new_param['type'] = splitted_params[0]
        parameters[name] = new_param
    for name, pattern in re.findall('\\(\\?P<([^>]+)>([^)]+)\\)', regex):
        if name not in parameters.keys():
            parameters[name] = {'name': name}
        parameters[name]['pattern'] = pattern
    return list(parameters.values())


def get_identifier(method, openapi_path):
    """Format the identifier shared in common with the endpoints and the transactions."""
    method = method.lower()
    return f'{method.lower()}<>{openapi_path}'



"""The formatter that will be passed to loguru."""


def format_logs(record):
    """Format loguru records properly."""
    if record['level'].name == 'SUCCESS':
        return """<bold><blue>escape</blue></bold> | <bold><green>{level}</green></bold> | <green>{message}</green>
"""
    if record['level'].name == 'WARNING':
        return """<bold><blue>escape</blue></bold> | <bold><yellow>{level}</yellow></bold> | <yellow>{message}</yellow>
"""
    if record['level'].name == 'ERROR':
        return """<bold><blue>escape</blue></bold> | <bold><red>{level}</red></bold> | <red>{message}</red>
"""
    return (
        '<bold><blue>escape</blue></bold> | <bold><blue>{level}</blue></bold> | {message}\n'
        )



"""Compute the converage at the end of the process."""
import os
import re
from datetime import datetime
from termcolor import cprint


def compute_coverage_from_transactions(transactions, endpoints, options):
    """This function is responsible for computing the raw coverage object This is a list of descriptors with the following structure :



    - path -> the path with the format open api

    - method -> the method of the endpoint

    - statuses -> the list of statuses covered by tests for this endpoint

    - covered -> a boolean describing if this route is covered by at least one test

    - filtered -> a boolean describing if the route is filtered by user options

    """
    coverage = {}
    ignored_paths = options.get('ignoredPaths', [])

    def is_filtered(path):
        """Determined if coverage result is filtered or not."""
        filtered = False
        for ignored_path in ignored_paths:
            matches = re.findall(re.compile(ignored_path), path)
            if len(matches) > 0:
                filtered = True
                break
        return filtered
    for endpoint in endpoints:
        statuses = [transaction['res']['statusCode'] for transaction in
            transactions if transaction['_identifier'] == endpoint[
            '_identifier']]
        statuses = list(dict.fromkeys(statuses))
        coverage[endpoint['_identifier']] = {'path': endpoint['openApiPath'
            ], 'method': endpoint['method'], 'statuses': statuses,
            'covered': bool(statuses), 'filtered': is_filtered(endpoint[
            'openApiPath'])}
    return coverage


def serialize_coverage(raw_coverage):
    """Prepare the coverage result for reporting.



    - compute the coverage rate

    """
    raw_coverage_list = raw_coverage.values()
    filtered_coverage = list(filter(lambda entry: entry['filtered'] is 
        False, raw_coverage_list))
    filtered_coverage.sort(key=lambda entry: entry['path'])
    n_endpoints_covered = 0
    n_endpoints = len(filtered_coverage)
    for entry in filtered_coverage:
        if entry['covered']:
            n_endpoints_covered += 1
    coverage_stats = {'n_endpoints': n_endpoints, 'n_endpoints_covered':
        n_endpoints_covered, 'coverage_rate': n_endpoints_covered / n_endpoints
        }
    return filtered_coverage, coverage_stats


def _center_between(text, padding_char, n_columns):
    """Util for displaying a centered text."""
    border_size = (n_columns - len(text)) // 2
    result = padding_char * border_size + text + padding_char * border_size
    if len(result) < n_columns:
        result += padding_char
    return padding_char * border_size + text + padding_char * border_size


def display_coverage_reports(data, stats, options):
    """Display the reports of the coverage based on reporters specified in options."""
    n_columns = os.get_terminal_size().columns
    sep = '=' * n_columns
    global_color = 'green'
    should_raise = False
    threshold = options.get('threshold', None)
    reporters = options.get('reporters', ['text', 'text-summary'])
    coverage_percentage = stats['coverage_rate'] * 100
    if threshold:
        below_coverage = coverage_percentage < threshold
        global_color = 'red' if below_coverage else 'green'
        should_raise = below_coverage
    cprint(_center_between(' Coverage report provided by Escape ', '=',
        n_columns), global_color, attrs=['bold'])
    cprint(f'Generated on : {datetime.now().isoformat()}', 'white', attrs=[
        'bold'])
    cprint(sep, global_color, attrs=['bold'])
    if 'text' in reporters:
        print()
        current_fragment = None

        def eventually_update_current_fragment(entry_path, current_fragment
            =None):
            """Print a line separator for root path changes in the report."""
            fragments = entry_path.split('/')
            print(entry_path)
            fragments = fragments or ['']
            if fragments[0] != current_fragment:
                current_fragment = fragments[0]
                section_title = f' /{current_fragment} '
                cprint('-' * 5 + section_title + '-' * (n_columns - 5 - len
                    (section_title)), 'white')
            return current_fragment
        for entry in data:
            current_fragment = eventually_update_current_fragment(entry[
                'path'], current_fragment)
            color = 'green' if entry['covered'] else 'red'
            cprint(
                f"{entry['method']} {entry['path']} : {', '.join(str(status) for status in entry['statuses'])}"
                , color)
        print()
    if 'text-summary' in reporters:
        cprint('=' * 5 + ' Summary ' + '=' * (n_columns - 5 - len(
            ' Summary ')), global_color, attrs=['bold'])
        cprint(
            f"Routes covered : {stats['n_endpoints_covered']}/{stats['n_endpoints']} => {round(coverage_percentage, 2)}%"
            , global_color)
        cprint('Per status :', global_color)
        cprint(sep, global_color, attrs=['bold'])
    if should_raise:
        raise ValueError(
            f'The route coverage is below expectation of {threshold}%. Received {round(coverage_percentage, 2)}%'
            )


def run_coverage(transactions, endpoints, config):
    """Execute coverage computation and serialization."""
    coverage_result = compute_coverage_from_transactions(transactions,
        endpoints, config)
    filtered_coverage, coverage_stats = serialize_coverage(coverage_result)
    enriched_endpoints = endpoints.copy()
    for enpoint in enriched_endpoints:
        enpoint['coverage'] = coverage_result[enpoint['_identifier']]
    return filtered_coverage, coverage_stats, enriched_endpoints



"""Django middleware."""
import os
import json
from loguru import logger
from django.http import HttpRequest, HttpResponse
from django.urls import get_resolver, URLPattern, URLResolver
from escape_cli.static.constants import METHODS_PATH, ENDPOINTS_PATH
from escape_cli.utils.result import save_transaction
from escape_cli.utils.parse import find_params, get_identifier, parse_body, parse_parameters, format_lib_to_openapi_path, interpolate_openapi_path
HTTP_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'PATCH']


class GetRequestAndResponseInformation:
    """Fetch usefull information from requests and responses Save it in the transactions file."""

    def __init__(self, get_response):
        self.result = {}
        self.get_response = get_response
        save_endpoints()

    def __call__(self, request):
        """Fetch the response after the view is called."""
        response = self.get_response(request)
        if 'req' not in self.result:
            self.result['req'] = {'openApiPath': '', 'protocol': request.
                scheme, 'host': request.get_host(), 'route': request.path,
                'method': request.method, 'parameters': {}, 'query': dict(
                request.GET.items()), 'originalUrl': request.path,
                'headers': dict(request.headers), 'cookies': request.
                COOKIES, 'httpVersion': request.META['SERVER_PROTOCOL'].
                split('/')[1], 'body': parse_body(request.body, request.
                content_type)}
        match = request.resolver_match
        if match:
            self.result['req']['openApiPath'] = format_lib_to_openapi_path(
                match.route)
        else:
            logger.warning(
                f'Cannot resolve route of {request.path}, interpolate openApiPath.'
                )
            self.result['req']['openApiPath'] = interpolate_openapi_path(self
                .result['req']['originalUrl'], self.result['req']['parameters']
                )
        self.result['_identifier'] = get_identifier(self.result['req'][
            'method'], self.result['req']['openApiPath'])
        res_headers = dict(response.items())
        res_body = parse_body(response.getvalue(), res_headers.get(
            'Content-Type', ''))
        self.result['res'] = {'statusCode': response.status_code,
            'messageCode': response.reason_phrase, 'headers': dict(response
            .items()), 'body': res_body}
        save_transaction(self.result)
        return response

    def process_view(self, request, _view_func, _view_args, view_kwargs):
        """Fetch the request before the view is called."""
        req_headers = dict(request.headers)
        req_body = parse_body(request.body, request.content_type)
        req_parameters = parse_parameters(view_kwargs)
        self.result['req'] = {'openApiPath': '', 'protocol': request.scheme,
            'host': request.get_host(), 'route': request.path, 'method':
            request.method, 'parameters': req_parameters, 'query': dict(
            request.GET.items()), 'originalUrl': request.path, 'headers':
            req_headers, 'cookies': request.COOKIES, 'httpVersion': request
            .META['SERVER_PROTOCOL'].split('/')[1], 'body': req_body}


def save_endpoints():
    """Save ground truth endpoints."""
    if os.environ['ESCAPE_ENDPOINTS_MAPPED'] == 'False':
        os.environ['ESCAPE_ENDPOINTS_MAPPED'] = 'True'
        endpoints = []
        endpoints = recurse_endpoints(endpoints, [get_resolver()])
        with open(ENDPOINTS_PATH, 'w') as f:
            json.dump(endpoints, f)


def recurse_endpoints(endpoints, resolvers, endpoint_start='', parameters=
    None, existing_routes=None):
    """Recursive function.



    Look throw the urls and save the endpoints.

    """
    existing_routes = existing_routes or []
    parameters = parameters or []
    with open(METHODS_PATH) as f:
        methods_dict = json.load(f)
    for url in resolvers:
        new_parameters = parameters + find_params(str(url.pattern), url.
            pattern.regex.pattern)
        lib_path = f'{endpoint_start}{url.pattern}'.lstrip('^')
        if isinstance(url, URLResolver):
            endpoints = recurse_endpoints(endpoints, url.url_patterns,
                lib_path, new_parameters, existing_routes)
        elif isinstance(url, URLPattern):
            openapi_path = format_lib_to_openapi_path(lib_path)
            if openapi_path not in existing_routes:
                existing_routes.append(openapi_path)
                handler = f'{url.callback.__module__}.{url.callback.__name__}'
                allowed_methods = HTTP_METHODS
                django_decorated = False
                if handler in methods_dict:
                    allowed_methods = methods_dict[handler]
                    django_decorated = True
                for method in allowed_methods:
                    endpoints.append({'openApiPath': openapi_path,
                        'libPath': lib_path, 'method': method, 'parameters':
                        new_parameters, 'handler': handler,
                        'djangoDecorated': django_decorated, '_identifier':
                        get_identifier(method, openapi_path)})
    return endpoints
